part of datepicker;

typedef _UpdatePickerState = void Function(
    _PickerStateArgs _updatePickerStateDetails);

/// Returns current view visible date range with current picker view.
typedef DateRangePickerViewChangedCallback = void Function(
    DateRangePickerViewChangedArgs);

/// Returns the old and new values of selection.
typedef DateRangePickerSelectionChangedCallback = void Function(
    DateRangePickerSelectionChangedArgs);

// method that raise the picker selection changed call back with the given parameters
void _raiseSelectionChangedCallback(SfDateRangePicker picker, {dynamic value}) {
  if (picker.onSelectionChanged == null) {
    return;
  }

  picker.onSelectionChanged(DateRangePickerSelectionChangedArgs(value));
}

// method that raises the visible dates changed call back with the given parameters
void _raisePickerViewChangedCallback(SfDateRangePicker picker,
    {PickerDateRange visibleDateRange, DateRangePickerView view}) {
  if (picker.onViewChanged == null) {
    return;
  }

  picker.onViewChanged(DateRangePickerViewChangedArgs(visibleDateRange, view));
}

class SfDateRangePicker extends StatefulWidget {
  SfDateRangePicker({
    Key key,
    DateRangePickerView view,
    this.selectionMode = DateRangePickerSelectionMode.single,
    this.headerHeight = 40,
    this.todayHighlightColor,
    this.backgroundColor,
    DateTime initialSelectedDate,
    List<DateTime> initialSelectedDates,
    PickerDateRange initialSelectedRange,
    List<PickerDateRange> initialSelectedRanges,
    this.toggleDaySelection = false,
    this.enablePastDates = true,
    this.showNavigationArrow = false,
    this.selectionShape = DateRangePickerSelectionShape.circle,
    this.navigationDirection = DateRangePickerNavigationDirection.horizontal,
    this.controller,
    this.onViewChanged,
    this.onSelectionChanged,
    DateRangePickerHeaderStyle headerStyle,
    DateRangePickerYearCellStyle yearCellStyle,
    DateRangePickerMonthViewSettings monthViewSettings,
    DateTime initialDisplayDate,
    DateTime minDate,
    DateTime maxDate,
    DateRangePickerMonthCellStyle monthCellStyle,
    this.monthFormat,
  })  : headerStyle = headerStyle ?? DateRangePickerHeaderStyle(),
        yearCellStyle = yearCellStyle ?? DateRangePickerYearCellStyle(),
        initialSelectedDate =
            controller != null && controller.selectedDate != null
                ? controller.selectedDate
                : initialSelectedDate,
        initialSelectedDates =
            controller != null && controller.selectedDates != null
                ? controller.selectedDates
                : initialSelectedDates,
        initialSelectedRange =
            controller != null && controller.selectedRange != null
                ? controller.selectedRange
                : initialSelectedRange,
        initialSelectedRanges =
            controller != null && controller.selectedRanges != null
                ? controller.selectedRanges
                : initialSelectedRanges,
        view = controller != null && controller.view != null
            ? controller.view
            : (view ?? DateRangePickerView.month),
        monthViewSettings =
            monthViewSettings ?? DateRangePickerMonthViewSettings(),
        initialDisplayDate =
            controller != null && controller.displayDate != null
                ? controller.displayDate
                : (initialDisplayDate ??
                    DateTime(DateTime.now().year, DateTime.now().month,
                        DateTime.now().day, 08, 45, 0)),
        minDate = minDate ?? DateTime(1900, 01, 01),
        maxDate = maxDate ?? DateTime(2100, 12, 31),
        monthCellStyle = monthCellStyle ?? DateRangePickerMonthCellStyle(),
        super(key: key);

  final DateRangePickerView view;

  final DateRangePickerSelectionMode selectionMode;

  final DateRangePickerHeaderStyle headerStyle;

  final double headerHeight;

  final Color todayHighlightColor;

  final Color backgroundColor;

  final bool toggleDaySelection;

  final DateRangePickerMonthViewSettings monthViewSettings;

  final DateRangePickerYearCellStyle yearCellStyle;

  final DateRangePickerMonthCellStyle monthCellStyle;

  final DateTime initialDisplayDate;

  final DateTime initialSelectedDate;

  final DateTime minDate;

  final DateTime maxDate;

  final bool enablePastDates;

  final List<DateTime> initialSelectedDates;

  final PickerDateRange initialSelectedRange;

  final List<PickerDateRange> initialSelectedRanges;

  final DateRangePickerController controller;

  final bool showNavigationArrow;

  final DateRangePickerNavigationDirection navigationDirection;

  final DateRangePickerSelectionShape selectionShape;

  final String monthFormat;

  final DateRangePickerViewChangedCallback onViewChanged;

  final DateRangePickerSelectionChangedCallback onSelectionChanged;

  @override
  _SfDateRangePickerState createState() => _SfDateRangePickerState();
}

class _SfDateRangePickerState extends State<SfDateRangePicker> {
  List<DateTime> _currentViewVisibleDates;
  DateTime _currentDate, _selectedDate;
  double _minWidth, _minHeight;
  ValueNotifier<List<DateTime>> _headerVisibleDates;
  List<DateTime> _selectedDates;
  PickerDateRange _selectedRange;
  List<PickerDateRange> _selectedRanges;
  dynamic _scrollViewKey;
  DateRangePickerView _view;
  bool _isRtl;
  DateRangePickerController _controller;
  Locale _locale;
  SfDateRangePickerThemeData _datePickerTheme;

  @override
  void initState() {
    _isRtl = false;
    _scrollViewKey = GlobalKey<_PickerScrollViewState>();
    //// Update initial values to controller.
    _initPickerController();
    _initNavigation();
    //// Update initial value to state variables.
    _updateSelectionValues();
    _view = _controller.view;
    _updateCurrentVisibleDates();
    _headerVisibleDates =
        ValueNotifier<List<DateTime>>(_currentViewVisibleDates);
    _controller.addPropertyChangedListener(_pickerValueChangedListener);
    super.initState();
  }

  void _initNavigation() {
    _controller._forward = _moveToNextView;
    _controller._backward = _moveToPreviousView;
  }

  void _initPickerController() {
    _controller = widget.controller ?? DateRangePickerController();
    _controller.selectedDate = widget.initialSelectedDate;
    _controller.selectedDates = _cloneList(widget.initialSelectedDates);
    _controller.selectedRange = widget.initialSelectedRange;
    _controller.selectedRanges = _cloneList(widget.initialSelectedRanges);
    _controller.view = widget.view;
    _currentDate =
        getValidDate(widget.minDate, widget.maxDate, widget.initialDisplayDate);
    _controller.displayDate = _currentDate;
  }

  void _updateSelectionValues() {
    _selectedDate = _controller.selectedDate;
    _selectedDates = _cloneList(_controller.selectedDates);
    _selectedRange = _controller.selectedRange;
    _selectedRanges = _cloneList(_controller.selectedRanges);
  }

  @override
  void dispose() {
    _controller.removePropertyChangedListener(_pickerValueChangedListener);
    _controller._backward = null;
    _controller._forward = null;
    super.dispose();
  }

  void _pickerValueChangedListener(String value) {
    if (value == _kSelectedDateString) {
      _raiseSelectionChangedCallback(widget, value: _controller.selectedDate);
      if (!mounted || isSameDate(_selectedDate, _controller.selectedDate)) {
        return;
      }

      setState(() {
        _selectedDate = _controller.selectedDate;
      });
    } else if (value == _kSelectedDatesString) {
      _raiseSelectionChangedCallback(widget, value: _controller.selectedDates);
      if (!mounted ||
          _isDateCollectionEquals(_selectedDates, _controller.selectedDates)) {
        return;
      }

      setState(() {
        _selectedDates = _cloneList(_controller.selectedDates);
      });
    } else if (value == _kSelectedRangeString) {
      _raiseSelectionChangedCallback(widget, value: _controller.selectedRange);
      if (!mounted ||
          _isRangeEquals(_selectedRange, _controller.selectedRange)) {
        return;
      }

      setState(() {
        _selectedRange = _controller.selectedRange;
      });
    } else if (value == _kSelectedRangesString) {
      _raiseSelectionChangedCallback(widget, value: _controller.selectedRanges);
      if (!mounted ||
          _isDateRangesEquals(_selectedRanges, _controller.selectedRanges)) {
        return;
      }

      setState(() {
        _selectedRanges = _cloneList(_controller.selectedRanges);
      });
    } else if (value == _kViewString) {
      if (!mounted || _view == _controller.view) {
        return;
      }

      setState(() {
        _view = _controller.view;
        _scrollViewKey.currentState._position = 0.0;
        _scrollViewKey.currentState._children.clear();
        _scrollViewKey.currentState._updateVisibleDates();
      });
    } else if (value == _kDisplayDateString) {
      if (!isSameOrAfterDate(widget.minDate, _controller.displayDate)) {
        _controller.displayDate = widget.minDate;
        return;
      }

      if (!isSameOrBeforeDate(widget.maxDate, _controller.displayDate)) {
        _controller.displayDate = widget.maxDate;
        return;
      }

      //// Restrict the update when current visible dates holds updated display date.
      if (isSameDate(_currentDate, _controller.displayDate) ||
          _checkDateWithInVisibleDates(_controller.displayDate)) {
        _currentDate = _controller.displayDate;
        return;
      }

      if (!mounted) {
        return;
      }

      setState(() {
        _currentDate = _controller.displayDate;
        _updateCurrentVisibleDates();
      });
    }
  }

  bool _checkDateWithInVisibleDates(DateTime date) {
    if (_controller.view == DateRangePickerView.month) {
      if (widget.monthViewSettings.numberOfWeeksInView != 6) {
        return isDateWithInDateRange(
            _currentViewVisibleDates[0],
            _currentViewVisibleDates[_currentViewVisibleDates.length - 1],
            date);
      } else {
        final DateTime _currentMonth =
            _currentViewVisibleDates[_currentViewVisibleDates.length ~/ 2];
        return date.month == _currentMonth.month &&
            date.year == _currentMonth.year;
      }
    } else if (_controller.view == DateRangePickerView.year) {
      final int _currentYear = _currentViewVisibleDates[0].year;
      final int _year = date.year;
      return _currentYear == _year;
    } else if (_controller.view == DateRangePickerView.decade) {
      final int _minYear = _currentViewVisibleDates[0].year;
      final int _maxYear = _currentViewVisibleDates[10].year - 1;
      final int _year = date.year;
      return _minYear <= _year && _maxYear >= _year;
    } else if (_controller.view == DateRangePickerView.century) {
      final int _minYear = _currentViewVisibleDates[0].year;
      final int _maxYear = _currentViewVisibleDates[10].year - 1;
      final int _year = date.year;
      return _minYear <= _year && _maxYear >= _year;
    }

    return false;
  }

  void _updateCurrentVisibleDates() {
    if (_view == DateRangePickerView.month) {
      _currentViewVisibleDates = getVisibleDates(
          _currentDate,
          null,
          widget.monthViewSettings.firstDayOfWeek,
          _getViewDatesCount(
              _view, widget.monthViewSettings.numberOfWeeksInView));
    } else {
      _currentViewVisibleDates = _getVisibleYearDates(_currentDate, _view);
    }
  }

  @override
  void didUpdateWidget(SfDateRangePicker oldWidget) {
    if (oldWidget.controller != widget.controller) {
      oldWidget.controller
          ?.removePropertyChangedListener(_pickerValueChangedListener);
      if (widget.controller != null) {
        _controller.selectedDate = widget.controller.selectedDate;
        _controller.selectedDates = _cloneList(widget.controller.selectedDates);
        _controller.selectedRange = widget.controller.selectedRange;
        _controller.selectedRanges =
            _cloneList(widget.controller.selectedRanges);
        _controller.view = widget.controller.view;
        _controller.displayDate = widget.controller.displayDate ?? _currentDate;
        _currentDate = getValidDate(
            widget.minDate, widget.maxDate, _controller.displayDate);
      } else {
        _initPickerController();
      }

      _controller.view ??= _view;
      _controller.addPropertyChangedListener(_pickerValueChangedListener);
      _initNavigation();
      _updateSelectionValues();
      _view = _controller.view;
    }

    if (oldWidget.selectionMode != widget.selectionMode) {
      _updateSelectionValues();
    }

    if (oldWidget.minDate != widget.minDate ||
        oldWidget.maxDate != widget.maxDate) {
      _currentDate = getValidDate(widget.minDate, widget.maxDate, _currentDate);
    }

    if (widget.monthViewSettings.numberOfWeeksInView !=
        oldWidget.monthViewSettings.numberOfWeeksInView) {
      _currentDate = _updateCurrentDate(oldWidget);
      _controller.displayDate = _currentDate;
    }

    if (oldWidget.controller != widget.controller ||
        widget.controller == null) {
      super.didUpdateWidget(oldWidget);
      return;
    }

    if (oldWidget.controller.selectedDate != widget.controller.selectedDate) {
      _selectedDate = _controller.selectedDate;
    }

    if (oldWidget.controller.selectedDates != widget.controller.selectedDates) {
      _selectedDates = _cloneList(_controller.selectedDates);
    }

    if (oldWidget.controller.selectedRange != widget.controller.selectedRange) {
      _selectedRange = _controller.selectedRange;
    }

    if (oldWidget.controller.selectedRanges !=
        widget.controller.selectedRanges) {
      _selectedRanges = _cloneList(_controller.selectedRanges);
    }

    if (oldWidget.controller.view != widget.controller.view) {
      _view = _controller.view;
      _currentDate = _updateCurrentDate(oldWidget);
      _controller.displayDate = _currentDate;
    }

    if (oldWidget.controller.displayDate != widget.controller.displayDate &&
        widget.controller.displayDate != null) {
      _currentDate =
          getValidDate(widget.minDate, widget.maxDate, _controller.displayDate);
      _controller.displayDate = _currentDate;
    }

    super.didUpdateWidget(oldWidget);
  }

  DateTime _updateCurrentDate(SfDateRangePicker _oldWidget) {
    if (_oldWidget.controller == widget.controller &&
        widget.controller != null &&
        _oldWidget.controller.view == DateRangePickerView.month &&
        _controller.view != DateRangePickerView.month) {
      return _currentViewVisibleDates[_currentViewVisibleDates.length ~/ 2];
    }

    return _currentViewVisibleDates[0];
  }

  @override
  void didChangeDependencies() {
    final TextDirection direction = Directionality.of(context);
    // default width value will be device width when the widget placed inside a infinity width widget
    _minWidth = MediaQuery.of(context).size.width;
    // default height for the widget when the widget placed inside a infinity height widget
    _minHeight = 300;
    _locale = Localizations.localeOf(context);
    _datePickerTheme = SfDateRangePickerTheme.of(context);
    _isRtl = direction != null && direction == TextDirection.rtl;
    super.didChangeDependencies();
  }

  @override
  Widget build(BuildContext context) {
    SyncfusionLicense.validateLicense(context);
    double _top = 0, _height;
    return LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
      _minWidth = constraints.maxWidth == double.infinity
          ? _minWidth
          : constraints.maxWidth;
      _minHeight = constraints.maxHeight == double.infinity
          ? _minHeight
          : constraints.maxHeight;

      _height = _minHeight - widget.headerHeight;
      _top = widget.headerHeight;
      if (_view == DateRangePickerView.month &&
          widget.navigationDirection ==
              DateRangePickerNavigationDirection.vertical) {
        _height -= widget.monthViewSettings.viewHeaderHeight;
        _top += widget.monthViewSettings.viewHeaderHeight;
      }

      return Container(
        width: _minWidth,
        height: _minHeight,
        color: widget.backgroundColor ?? _datePickerTheme.backgroundColor,
        child: _addChildren(_top, _height, _minWidth),
      );
    });
  }

  Widget _addChildren(double _top, double _height, double _width) {
    _headerVisibleDates.value = _currentViewVisibleDates;
    return Stack(children: <Widget>[
      Positioned(
        top: 0,
        right: 0,
        left: 0,
        height: widget.headerHeight,
        child: GestureDetector(
          child: Container(
            color: widget.headerStyle.backgroundColor ??
                _datePickerTheme.headerBackgroundColor,
            child: _PickerHeaderView(
                _headerVisibleDates,
                widget.headerStyle,
                widget.selectionMode,
                _view,
                widget.monthViewSettings.numberOfWeeksInView,
                widget.showNavigationArrow,
                widget.navigationDirection,
                widget.monthViewSettings.enableSwipeSelection,
                widget.minDate,
                widget.maxDate,
                widget.monthFormat,
                _datePickerTheme,
                _locale,
                _width,
                widget.headerHeight,
                _moveToPreviousView,
                _moveToNextView,
                _isRtl),
            height: widget.headerHeight,
          ),
          onTapUp: (TapUpDetails details) {
            _updateCalendarTapCallbackForHeader();
          },
        ),
      ),
      _getViewHeaderView(),
      Positioned(
        top: _top,
        left: 0,
        right: 0,
        height: _height,
        child: _PickerScrollView(
          widget,
          _controller,
          _width,
          _height,
          _isRtl,
          _datePickerTheme,
          _locale,
          getPickerStateValues: (_PickerStateArgs details) {
            _getPickerStateValues(details);
          },
          updatePickerStateValues: (_PickerStateArgs details) {
            _updatePickerStateValues(details);
          },
          key: _scrollViewKey,
        ),
      ),
    ]);
  }

  Widget _getViewHeaderView() {
    if (_view == DateRangePickerView.month &&
        widget.navigationDirection ==
            DateRangePickerNavigationDirection.vertical) {
      final Color todayTextColor =
          widget.monthCellStyle.todayTextStyle != null &&
                  widget.monthCellStyle.todayTextStyle.color != null
              ? widget.monthCellStyle.todayTextStyle.color
              : (widget.todayHighlightColor != null &&
                      widget.todayHighlightColor != Colors.transparent
                  ? widget.todayHighlightColor
                  : Colors.blue);
      return Positioned(
        left: 0,
        top: widget.headerHeight,
        right: 0,
        height: widget.monthViewSettings.viewHeaderHeight,
        child: Container(
          color: widget.monthViewSettings.viewHeaderStyle.backgroundColor ??
              _datePickerTheme.viewHeaderBackgroundColor,
          child: RepaintBoundary(
            child: CustomPaint(
              painter: _PickerViewHeaderPainter(
                  _currentViewVisibleDates,
                  widget.monthViewSettings.viewHeaderStyle,
                  widget.monthViewSettings.viewHeaderHeight,
                  widget.monthViewSettings,
                  _datePickerTheme,
                  _locale,
                  _isRtl,
                  widget.monthCellStyle,
                  todayTextColor),
            ),
          ),
        ),
      );
    }

    return Positioned(left: 0, top: 0, right: 0, height: 0, child: Container());
  }

  void _moveToNextView() {
    if (!_canMoveToNextView(_view, widget.monthViewSettings.numberOfWeeksInView,
        widget.maxDate, _currentViewVisibleDates)) {
      return;
    }

    _isRtl
        ? _scrollViewKey.currentState._moveToPreviousViewWithAnimation()
        : _scrollViewKey.currentState._moveToNextViewWithAnimation();
  }

  void _moveToPreviousView() {
    if (!_canMoveToPreviousView(
        _view,
        widget.monthViewSettings.numberOfWeeksInView,
        widget.minDate,
        _currentViewVisibleDates)) {
      return;
    }

    _isRtl
        ? _scrollViewKey.currentState._moveToNextViewWithAnimation()
        : _scrollViewKey.currentState._moveToPreviousViewWithAnimation();
  }

  dynamic _getPickerStateValues(_PickerStateArgs details) {
    details._currentDate = details._currentDate ?? _currentDate;
    details._selectedDate = _selectedDate;
    details._selectedDates = _selectedDates;
    details._selectedRange = _selectedRange;
    details._selectedRanges = _selectedRanges;
    details._view = _view;
  }

  dynamic _updatePickerStateValues(_PickerStateArgs details) {
    if (details._currentDate != null) {
      if (!isSameOrAfterDate(widget.minDate, details._currentDate)) {
        details._currentDate = widget.minDate;
      }

      if (!isSameOrBeforeDate(widget.maxDate, details._currentDate)) {
        details._currentDate = widget.maxDate;
      }

      _currentDate = details._currentDate;
      _controller.displayDate = _currentDate;
    }

    if (details._currentViewVisibleDates != null &&
        _currentViewVisibleDates != details._currentViewVisibleDates) {
      _currentViewVisibleDates = details._currentViewVisibleDates;
      _headerVisibleDates.value = _currentViewVisibleDates;
      if (!widget.monthViewSettings.showTrailingAndLeadingDates &&
          widget.monthViewSettings.numberOfWeeksInView == 6 &&
          _controller.view == DateRangePickerView.month) {
        final DateTime _visibleDate =
            _currentViewVisibleDates[_currentViewVisibleDates.length ~/ 2];
        _raisePickerViewChangedCallback(widget,
            visibleDateRange: PickerDateRange(_getMonthStartDate(_visibleDate),
                _getMonthEndDate(_visibleDate)),
            view: _controller.view);
      } else {
        _raisePickerViewChangedCallback(widget,
            visibleDateRange: PickerDateRange(_currentViewVisibleDates[0],
                _currentViewVisibleDates[_currentViewVisibleDates.length - 1]),
            view: _controller.view);
      }
    }

    if (details._view != null && _view != details._view) {
      _controller.view = details._view;
    }

    if (_view == DateRangePickerView.month) {
      if (widget.selectionMode == DateRangePickerSelectionMode.single) {
        _selectedDate = details._selectedDate;
        _controller.selectedDate = _selectedDate;
      } else if (widget.selectionMode ==
          DateRangePickerSelectionMode.multiple) {
        _selectedDates = details._selectedDates;
        _controller.selectedDates = _selectedDates;
      } else if (widget.selectionMode == DateRangePickerSelectionMode.range) {
        _selectedRange = details._selectedRange;
        _controller.selectedRange = _selectedRange;
      } else if (widget.selectionMode ==
          DateRangePickerSelectionMode.multiRange) {
        _selectedRanges = details._selectedRanges;
        _controller.selectedRanges = _selectedRanges;
      }
    }
  }

  // method to update the calendar tapped call back for the header view
  void _updateCalendarTapCallbackForHeader() {
    if (_view == DateRangePickerView.century) {
      return;
    }

    if (_view == DateRangePickerView.month) {
      _controller.view = DateRangePickerView.year;
    } else {
      if (_view == DateRangePickerView.year) {
        _controller.view = DateRangePickerView.decade;
      } else if (_view == DateRangePickerView.decade) {
        _controller.view = DateRangePickerView.century;
      }
    }
  }
}
