part of datepicker;

typedef DateRangePickerValueChangedCallback = void Function(String);

const String _kSelectedDateString = 'selectedDate';
const String _kSelectedDatesString = 'selectedDates';
const String _kSelectedRangeString = 'selectedRange';
const String _kSelectedRangesString = 'selectedRanges';
const String _kDisplayDateString = 'displayDate';
const String _kViewString = 'view';

class DateRangePickerValueChangeNotifier {
  List<DateRangePickerValueChangedCallback> _listeners;

  //// Add the listener used to call back when picker value changed.
  void addPropertyChangedListener(
      DateRangePickerValueChangedCallback _listener) {
    _listeners ??= <DateRangePickerValueChangedCallback>[];
    _listeners.add(_listener);
  }

  //// remove the listener used for notify the picker value changes.
  void removePropertyChangedListener(
      DateRangePickerValueChangedCallback _listener) {
    if (_listeners == null) {
      return;
    }

    _listeners.remove(_listener);
  }

  //// Notify to the listener while controller value changed.
  void notifyPropertyChangedListeners(String value) {
    if (_listeners == null) {
      return;
    }

    for (DateRangePickerValueChangedCallback listener in _listeners) {
      if (listener != null) {
        listener(value);
      }
    }
  }

  @mustCallSuper
  void dispose() {
    _listeners = null;
  }
}

class DateRangePickerController extends DateRangePickerValueChangeNotifier {
  DateTime _selectedDate;
  List<DateTime> _selectedDates;
  PickerDateRange _selectedRange;
  List<PickerDateRange> _selectedRanges;
  DateTime _displayDate;
  DateRangePickerView _view;

  DateTime get selectedDate => _selectedDate;

  set selectedDate(DateTime date) {
    if (isSameDate(_selectedDate, date)) {
      return;
    }

    _selectedDate = date;
    notifyPropertyChangedListeners(_kSelectedDateString);
  }

  List<DateTime> get selectedDates => _selectedDates;

  set selectedDates(List<DateTime> dates) {
    if (_isDateCollectionEquals(_selectedDates, dates)) {
      return;
    }

    _selectedDates = _cloneList(dates);
    notifyPropertyChangedListeners(_kSelectedDatesString);
  }

  PickerDateRange get selectedRange => _selectedRange;

  set selectedRange(PickerDateRange range) {
    if (_isRangeEquals(_selectedRange, range)) {
      return;
    }

    _selectedRange = range;
    notifyPropertyChangedListeners(_kSelectedRangeString);
  }

  List<PickerDateRange> get selectedRanges => _selectedRanges;

  set selectedRanges(List<PickerDateRange> ranges) {
    if (_isDateRangesEquals(_selectedRanges, ranges)) {
      return;
    }

    _selectedRanges = _cloneList(ranges);
    notifyPropertyChangedListeners(_kSelectedRangesString);
  }

  DateTime get displayDate => _displayDate;

  set displayDate(DateTime date) {
    if (isSameDate(_displayDate, date)) {
      return;
    }

    _displayDate = date;
    notifyPropertyChangedListeners(_kDisplayDateString);
  }

  DateRangePickerView get view => _view;

  set view(DateRangePickerView value) {
    if (_view == value) {
      return;
    }

    _view = value;
    notifyPropertyChangedListeners(_kViewString);
  }

  VoidCallback _forward;

  void forward() {
    if (_forward == null) {
      return;
    }

    _forward();
  }

  VoidCallback _backward;

  void backward() {
    if (_backward == null) {
      return;
    }

    _backward();
  }
}
